"use strict";
/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
const tfjs_core_1 = require("@tensorflow/tfjs-core");
const shader_preprocessor_1 = require("../shader_preprocessor");
const webgpu_util_1 = require("../webgpu_util");
class ArgMinMaxProgram {
    constructor(inputShape, axis, reduceType) {
        this.variableNames = ['x'];
        this.uniforms = 'uint axis;';
        const axes = [axis];
        tfjs_core_1.backend_util.assertAxesAreInnerMostDims('arg' + reduceType.charAt(0).toUpperCase() + reduceType.slice(1), axes, inputShape.length);
        const op = reduceType === 'min' ? '<' : '>';
        // |outShape| is the shape with the removed axis
        // |reduceShape| is the shape we are reducing. i.e. [ inputShape[axis] ]
        const [outputShape, reduceShape] = tfjs_core_1.backend_util.computeOutAndReduceShapes(inputShape, axes);
        this.outputShape = outputShape.length === 0 ? [1] : outputShape;
        // Length of the axis we're reducing on.
        const reduceSize = tfjs_core_1.util.sizeFromShape(reduceShape);
        // The number of comparisons each thread will do
        const reductionFactor = 2;
        const xMaxThreads = 1024; // gl_MaxComputeWorkGroupSize
        const xThreads = Math.min(Math.ceil(reduceSize / reductionFactor), xMaxThreads);
        this.workGroupSize = [xThreads, 1, 1];
        this.dispatchLayout = { x: [], y: this.outputShape.map((d, i) => i) };
        this.dispatch = webgpu_util_1.computeDispatch(this.dispatchLayout, this.outputShape);
        // When xThreads > 1, each thread reduces Length / xThreads values.
        // Thes results are stored in shared memory and iteratively reduced.
        const reduceInSharedMemory = xThreads > 1;
        const sharedMemorySnippet = `
      shared uint xBestIndices[WorkGroupSize];
      shared float xBestValues[WorkGroupSize];
    `;
        const sharedMemoryReduceSnippet = `
      xBestIndices[gl_LocalInvocationID.x] = bestIndex;
      xBestValues[gl_LocalInvocationID.x] = bestValue;

      uint currentSize = WorkGroupSize;
      while (currentSize > 1) {
        barrier();

        for (uint w = 0; w < ${reductionFactor}; ++w) {
          uint i = gl_LocalInvocationID.x * ${reductionFactor} + w;
          if (i < currentSize) {
            uint candidateIndex = xBestIndices[i];
            float candidate = xBestValues[i];
            if (candidate ${op} bestValue && !isnan(candidate)) {
              bestValue = candidate;
              bestIndex = candidateIndex;
            }
          }
        }

        xBestIndices[gl_LocalInvocationID.x] = bestIndex;
        xBestValues[gl_LocalInvocationID.x] = bestValue;

        currentSize = DIV_CEIL(currentSize, ${reductionFactor});
      }

      if (gl_LocalInvocationID.x == 0) {
        setOutput(flatOutputIndex, int(bestIndex));
      }
    `;
        const outputCoordsType = shader_preprocessor_1.getCoordsDataType(this.outputShape.length);
        const indexOutputCoords = (outputCoords, index) => {
            if (this.outputShape.length === 1) {
                return outputCoords;
            }
            else {
                return `${outputCoords}[${index}]`;
            }
        };
        const indexInputShape = (index) => {
            if (inputShape.length === 1) {
                return 'xShape';
            }
            else {
                return `xShape[${index}]`;
            }
        };
        this.userCode = `
      #define DIV_CEIL(x, y) (((x) - 1) / (y) + 1)

      const uint WorkGroupSize = gl_WorkGroupSize.x;

      ${reduceInSharedMemory ? sharedMemorySnippet : ''}

      // In order to get a flattened index into the input tensor, we need to
      // add back the index along the reduced dimension to |outputCoords|.
      // This function outputs the offset to the first value along
      // |axis| and the stride to get the next value of the input along |axis|.
      uvec2 getInputCoordInfo() {
        const ${outputCoordsType} outputCoords = getOutputCoords();
        uint i = ${this.outputShape.length - 1};

        uint stride = 1;
        uint inputStride = 1;
        uint offset = 0;

        for (uint r = 1; r <= ${inputShape.length}; ++r) {
          uint length = ${indexInputShape(`${inputShape.length} - r`)};
          if (${inputShape.length} - r == axis) {
            inputStride = stride;
          } else {
            offset += ${indexOutputCoords('outputCoords', 'i--')} * stride;
          }
          stride *= length;
        }

        return uvec2(offset, inputStride);
      }

      uint getInputIndex(uvec2 coordInfo, uint index) {
        return coordInfo[0] + coordInfo[1] * index;
      }

      void main() {
        const uvec2 coordInfo = getInputCoordInfo();

        uint bestIndex = 0;
        float bestValue = x[getInputIndex(coordInfo, bestIndex)];

        const uint Length = ${indexInputShape('axis')};
        const uint WorkPerThread = DIV_CEIL(Length, WorkGroupSize);

        for (uint w = 0; w < WorkPerThread; ++w) {
          uint i = gl_GlobalInvocationID.x * WorkPerThread + w;
          if (i < Length) {
            float candidate = x[getInputIndex(coordInfo, i)];
            if (candidate ${op} bestValue && !isnan(candidate)) {
              bestValue = candidate;
              bestIndex = i;
            }
          }
        }

        const uint flatOutputIndex = gl_GlobalInvocationID.y;
        ${reduceInSharedMemory ? sharedMemoryReduceSnippet :
            'setOutput(flatOutputIndex, int(bestIndex));'}
      }
    `;
    }
}
exports.ArgMinMaxProgram = ArgMinMaxProgram;
//# sourceMappingURL=argminmax_webgpu.js.map